/*
 * FILE --- rls_estm.cc
 *
 * PURPOSE --- Header file for a class that implements a Recursive 
 * Least Squares (RLS) algorithm for parameter estimation
 * for Single-Input-Single-Output (SISO) plants.
 * The user specifies the dimension of the parameter vector, the 
 * initial parameter vector and initial covariance matrix.
 * Moreover, the user can set a forgetting factor lambda, such that
 * 0 < lambda <= 1, where lambda = 1 is the no forgetting case.
 * 
 * This utility uses the matrix tool for matrix maltiplications.
 *
 */

// Include Files
#include "rls_estm.hh"

// Define this macro if you want to use this utility without linking
// to the RCS Library
#ifdef NO_RCS
#include <stdio.h>            /* printf function */
#else
#include "rcs_prnt.hh"          /* rcs_print_ functions */
#endif

/*************************************************************/

RLS_ESTM::RLS_ESTM(int _num_of_param, 
	 double *_theta_0 = NULL, 
	 double **_P_0 = NULL)
{
  // No error at start
  rls_estm_err = RLS_ESTM_NO_ERROR;
  num_of_param = _num_of_param;
  
  // Create the matrices
  theta = new matrix(num_of_param, 1);
  P = new matrix(num_of_param, num_of_param);
  h = new matrix(num_of_param, 1);
  Ph = new matrix(num_of_param, 1);
  
  // Check if matrixes are initialized correctly
  if ((theta->matrix_err != MATRIX_NO_ERROR) || 
      (P->matrix_err != MATRIX_NO_ERROR) ||
      (h->matrix_err != MATRIX_NO_ERROR) ||
      (Ph->matrix_err != MATRIX_NO_ERROR))
    {
      rls_estm_err = RLS_ESTM_MEM_ERROR;
    }
  
  // If any error then report
  if (rls_estm_err != RLS_ESTM_NO_ERROR)
    {
#ifdef NO_RCS
      printf("RLS_ESTM: ERROR --- Not Enought Memory!!!\n");
      printf("The RLS estimator will not function correctly ...\n");
#else
      rcs_print("RLS_ESTM: ERROR --- Not Enought Memory!!!\n");
      rcs_print("The RLS estimator will not function correctly ...\n");
#endif
    }

  // Initialize the values
  set_theta(_theta_0);
  set_P((const double **)_P_0);

  // Set no forgetting factor
  lambda = 1.0;

  // Initilaize the members
  output_error = 0.0;
  temp = 0.0;
}

RLS_ESTM::~RLS_ESTM()
{
  delete theta;
  delete P;
  delete h;
  delete Ph;
}

void RLS_ESTM::set_lambda(double _lambda)
{
  // Lambda should be b/n 0.0 and 1.0
  if (_lambda <= 0.0 || _lambda > 1.0)
    {
#ifdef NO_RCS
      printf("RLS_ESTM:set_lambda: ERROR --- Invalid lambda!!!\n");
      printf("The value should be 0.0 < lambda <= 1.0...\n");
#else
      rcs_print("RLS_ESTM:set_lambda: ERROR --- Invalid lambda!!!\n");
      rcs_print("The value should be 0.0 < lambda <= 1.0...\n");
#endif
    }
  else
    {
      lambda = _lambda;
    }
} 

void RLS_ESTM::set_theta(const double *_theta)
{
  if (_theta == NULL)
    {
      theta->set_col(0, 0.0);
    }
  else
    {
      theta->set_col(0, _theta);
    }
}

void RLS_ESTM::get_theta(double *_theta)
{
  // Return the parameters
  for(int ii = 0; ii < num_of_param; ii++)
    {
      _theta[ii] = theta->get_element(ii,0);
    }
}

void RLS_ESTM::set_P(const double **_P)
{
  if (_P == NULL)
    {
      P->set_matrix(0.0);
      P->set_diagonal(1.0);
    }
  else
    {
      P->set_matrix(_P);
    }
}

void RLS_ESTM::set_P(double _diagonal_val)
{
  P->set_matrix(0.0);
  P->set_diagonal(_diagonal_val);
}

void RLS_ESTM::get_P(double **_P)
{
  int ii, jj; // Index variables for the loops

  // Return the covariance matrix
  for(ii = 0; ii < num_of_param; ii++)
    {
      for(jj = 0; jj < num_of_param; jj++)
	{
	  _P[ii][jj] = P->get_element(ii,jj);
	}
    }
}

double RLS_ESTM::estimate(const double plant_output, 
		     const double *regression_vector,
		     double *estimates)
{
  // Use matrix operations
  h->set_col(0, regression_vector);

  // Calculate the error b/n the expected and actual autput
  output_error = plant_output - (transpose(*h) * (*theta)).get_element(0,0);

  // Covariance regression product (for conveniance)
  (*Ph) = (*P) * (*h);

  // The denominator (for conveniance)
  temp = lambda + (transpose(*h) * (*P) * (*h)).get_element(0,0);
 
  // Calculate the new estimate 
  (*theta) = (*theta) + (output_error/temp) * (*Ph);

  // Calculate the new covarinace P
  (*P) = (1.0/lambda) * ((*P) - (1.0/temp) * (*Ph) * transpose(*Ph));

  // Return the new estimates
  for(int ii = 0; ii < num_of_param; ii++)
    {
      estimates[ii] = theta->get_element(ii,0);
    }
  
  // Return the new output estimate
  return ( (transpose(*h) * (*theta)).get_element(0,0) );
}
  
